/*
############################################################################
# Copyright (c) 2010 - 2014 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __WATCOMC__
#include <direct.h>
#else
#include <dirent.h>
#endif
#include <setjmp.h>
#include <unistd.h>
#include "links.h"
#include "cline.h"
#include "vars.h"
#include "tokens.h"
#include "spinit.h"
#include "strsubs.h"
#include "cdirect.h"
#include "funcptr.h"

LinkT *FuncPtrList = 0;
extern VarInfoT *varlist;
extern VarInfoT *lvarlist;
extern TypeInfoT *typeslist;
extern ConInfoT *conlist;
extern LinkT *SymbolList;
extern int total_mem_alloc;
extern int total_mem_freed;
//extern int nonewline;
FILE *inputfile[10];
char inputfname[10][80];
int inputline[10];
int inputfileindex = 0;
FILE *outfile;
int whitespace = 1;
static int debugflag = 0;
static int spinflag = 1;
static int sourceflag = 0;
static int mainflag = 0;
static int interact = 0;
static int indent;
static int localflag;
static int charsign = 0; // Default char is unsigned

static char *versionstr = "0.82";

static LinkT *tokenlistout = 0;
static LinkT *tokenlistin = 0;
static LinkT *tokenlistdefer = 0;

void InitializeFuncProtoList(void);
void InitializeTypesList(void);
void InitializeSymbolList(void);
int ParseCommandLineArgs(int argc, char **argv);
int ProcessIncludeDirective(LinkT *tokenlistin);
void AddIntrinsicFunctions(void);
void PrintInitialInfo(void);

#if 0
void printtokenlistin(void)
{
    LinkT *link = LinkCreate("'", sizeof(LinkT));
    link->next = tokenlistin;
    printtokenlist(link, indent, 1);
    LinkDestroy(link);
}
#endif

void printtokenscomment(LinkT *tokenlist, int indent)
{
    LinkT *link = LinkCreate("'", sizeof(LinkT));
    link->next = tokenlist;
    printtokenlist(link, indent, 1);
    LinkDestroy(link);
}

int OpenFiles(char *fname)
{
    int len;
    char filename[100];

    strncpy(inputfname[0], fname, 79);
    inputfname[0][79] = 0;
    inputline[0] = 0;
    inputfile[0] = fopen(inputfname[0], "r");

    if (inputfile[0] == 0)
    {
        printf("Could not open input file %s\n", fname);
	return 0;
    }

    strcpy(filename, fname);
    len = strlen(filename);
    if (filename[len-2] == '.' && filename[len-1] == 'c') filename[len-2] = 0;
    strcat(filename, ".spin");
    outfile = fopen(filename, "w");

    if (outfile == 0)
    {
        printf("Could not open output file %s\n", filename);
	fclose (inputfile[0]);
	return 0;
    }

    return 1;
}

void DestroyAllLists(void)
{
    TypeInfoT *typeinfo = typeslist;

    // Destroy all the struct variable lists
    while (typeinfo)
    {
	if (typeinfo->structvars && typeinfo->typedefptr == 0)
	    ListDestroy(typeinfo->structvars);
	typeinfo = typeinfo->next;
    }

    // Destroy all the lists
    tokenlistin = ListDestroy(tokenlistin);
    tokenlistout = ListDestroy(tokenlistout);
    tokenlistdefer = ListDestroy(tokenlistdefer);
    varlist = (VarInfoT *)ListDestroy((LinkT *)varlist);
    lvarlist = (VarInfoT *)ListDestroy((LinkT *)lvarlist);
    conlist = DestroyConList(conlist);
    typeslist = (TypeInfoT *)ListDestroy((LinkT *)typeslist);
    SymbolList = ListDestroy(SymbolList);
}

static jmp_buf CspinExitBuf;

void CspinExit(int errval)
{
    if (errval == 0) errval = -1;
    longjmp(CspinExitBuf, errval);
}

typedef struct LineInfoS
{
    struct LineInfoS *next;
    int clinetype;
    LinkT *tokenlist;
    int newline;
    int oneliner;
} LineInfoT;

LineInfoT *AnalyzeTokenList(LinkT *tokenlist, int clinetype)
{
    char *str;
    int parencount;
    LineInfoT *line;
    LineInfoT *linelist =  0;

    while (tokenlist)
    {
        if (tokenlist->str[0] == '\n')
	{
            tokenlist = tokenlist->next;
	    continue;
	}

        line = (LineInfoT *)LinkCreate("", sizeof(LinkT));
        line->newline = 1;
        line->tokenlist = tokenlist;
	linelist = (LineInfoT *)ListAppend((LinkT *)linelist, (LinkT *)line);
        str = tokenlist->str;
        if (str[0] == '{')
	{
	    printf("CLINE_BRACE_LEFT\n");
	    line->clinetype = CLINE_BRACE_LEFT;
	    tokenlist = tokenlist->next;
	}
        else if (str[0] == '}')
	{
	    printf("CLINE_BRACE_RIGHT\n");
	    line->clinetype = CLINE_BRACE_RIGHT;
	}
        else if (strcmp(str, "if") == 0)
	{
            printf("CLINE_CONTROL\n");
            line->clinetype = CLINE_CONTROL;
	    parencount = 0;
	    while (tokenlist)
	    {
		if (tokenlist->str[0] == '(')
		    parencount++;
		else if (tokenlist->str[0] == ')')
		{
		     parencount--;
		     if (parencount <= 0) break;
		}
		tokenlist = tokenlist->next;
	    }
	}
        else if (strcmp(str, "for") == 0 || strcmp(str, "while") == 0)
	{
            printf("CLINE_LOOP\n");
            line->clinetype = CLINE_LOOP;
	    parencount = 0;
	    while (tokenlist)
	    {
		if (tokenlist->str[0] == '(')
		    parencount++;
		else if (tokenlist->str[0] == ')')
		{
		     parencount--;
		     if (parencount <= 0) break;
		}
		tokenlist = tokenlist->next;
	    }
	}
#if 0
	else if (strcmp(str, "typedef") == 0)
	{
	    while (tokenlist)
	    {
		if (tokenlist->str[0] == ';')
		{
		    printf("CLINE_TYPEDEF\n");
		    line->clinetype = CLINE_TYPEDEF;
		    tokenlist = tokenlist->next;
		    break;
		}
		else if (tokenlist->str[0] == '{')
		{
		    printf("CLINE_TYPE_STRUCT\n");
		    line->clinetype = CLINE_TYPE_STRUCT;
		    break;
		}
	        tokenlist = tokenlist->next;
	    }
	    continue;
	}
#define CLINE_FUNCTION       1
#define CLINE_DECL_LOCAL     3
#define CLINE_DECL_GLOBAL    4
#define CLINE_DIRECTIVE      7
#define CLINE_STATEMENT      8
#define CLINE_PROTOTYPE     10
#define CLINE_STRUCT        13
#define CLINE_STRUCT_TYPE   15
#define CLINE_STRUCT_VAR    16
#define CLINE_STRUCT_ELEM   17
#define CLINE_DECL_STRUCT   18
#define CLINE_SWITCH        21
#define CLINE_CASE          22
#define CLINE_SKIP          23
#define CLINE_COMMENT       32
#define CLINE_C_COMMENT     33
#define CLINE_C_COMMENTS    34
#endif
        else
	{
#if 0
	    printf("CLINE_STATEMENT\n");
            line->clinetype = CLINE_STATEMENT;
	    while (tokenlist)
	    {
		if (tokenlist->str[0] == ';') break;
		tokenlist = tokenlist->next;
	    }
#else
            printf("Unknown clinetype.  Look for newline\n");
	    line->clinetype = clinetype;
#if 0
	    if ((clinetype & CLINE_I_COMMENT) == CLINE_I_COMMENT)
                line->newline = 0;
#endif
	    while (tokenlist)
	    {
		if (tokenlist->str[0] == '\n') break;
		tokenlist = tokenlist->next;
	    }
#endif
	}
	if (tokenlist)
	{
            LinkT *next = tokenlist->next;
	    tokenlist->next = 0;
	    tokenlist = next;
	    if (tokenlist && tokenlist->str[0] == '\n')
	        tokenlist = tokenlist->next;
	}
    }

    return linelist;
}

void ProcessCLine(LineInfoT *linelist);

int ConvertFileToSpin(char *fname)
{
    int errval;
    int newline;
    int clinetype;
    int oneliner;
    LineInfoT linelist;

    indent = 0;
    localflag = 0;

    if ((errval = setjmp(CspinExitBuf))) return errval;

    inputfileindex = 0;
    if (OpenFiles(fname) == 0) return 0;

    initcline();
    InitDirect();
    InitSpinit();

    fprintf(outfile, "'*********************************************\n");
    fprintf(outfile, "'* Spin code generated by cspin version %s *\n", versionstr);
    fprintf(outfile, "'*********************************************\n");
    if (mainflag) PrintInitialInfo();
    InitializeTypesList();
    InitializeFuncProtoList();
    InitializeSymbolList();
    AddIntrinsicFunctions();

    while ((tokenlistin = getcline(&clinetype, &newline)))
    {
        indent = UpdateAndGetIndentLevel(clinetype, &oneliner);
	linelist.clinetype = clinetype;
	linelist.tokenlist = tokenlistin;
	linelist.newline = newline;
	linelist.oneliner = oneliner;
        ProcessCLine(&linelist);
        tokenlistin = ListDestroy(tokenlistin);
        tokenlistout = ListDestroy(tokenlistout);
    }

    if (FindVariable(varlist, "main"))
    {
        //fprintf(outfile, "\n' This program contains a main entry point\n");
    }

    fprintf(outfile, "\n");

    if (FuncPtrList)
    {
        if (whitespace)
            fprintf(outfile, "\n");
        fprintf(outfile, "OBJ\n");
        fprintf(outfile, "  mp : \"MethodPointer\"\n");
        if (whitespace)
            fprintf(outfile, "\n");
        fprintf(outfile, "DAT\n");
        PrintFuncPtrDat();
    }

    if (mainflag || FuncPtrList)
    {
        if (whitespace)
            fprintf(outfile, "\n");
        fprintf(outfile, "PUB cspin_initialize\n");
    }
    if (FuncPtrList)
    {
        fprintf(outfile, "  mp.Initialize\n");
        PrintFuncPtrInit();
    }

    if (debugflag)
    {
	 PrintTypes();
	 PrintConstants();
    }

    // Destroy all the list and free up the memory
    DestroyAllLists();

    if (debugflag)
    {
	 printf("\nTotal memory allocated     = %7d bytes\n", total_mem_alloc);
	 printf("Total memory freed         = %7d bytes\n", total_mem_freed);
	 printf("Remaining memory allocated = %7d bytes\n",
	     total_mem_alloc - total_mem_freed);
    }
    fprintf(outfile, "\n");
    fclose(outfile);
    return 0;
}

void ProcessCLine(LineInfoT *linelist)
{
    int clinetype;
    int newline;
    int oneliner;
    LinkT *tokenlist;

    if (linelist == 0) return;

    clinetype = linelist->clinetype;
    newline = linelist->newline;
    oneliner = linelist->oneliner;
    tokenlist = linelist->tokenlist;

    checkcline(tokenlist, clinetype);
    getcvars(tokenlist, clinetype);

    if (sourceflag && clinetype != CLINE_COMMENT &&
        clinetype != CLINE_C_COMMENT && clinetype != CLINE_C_COMMENTS)
        printtokenscomment(tokenlist, 2*indent);

    if (clinetype == CLINE_DECL_LOCAL)
    {
        localflag = 1;
        seperateinits(&tokenlist, &tokenlistdefer);
        if (!spinflag)
            printtokenlist(tokenlist, 4*indent, 1);
        else
        {
            ConvertToSpin(&tokenlist, &tokenlistout, clinetype);
	    tokenlistout = CheckTokenList(tokenlistout);
	    printtokenlist(tokenlistout, 2*indent, 0);
	    tokenlistout = ListDestroy(tokenlistout);
        }
    }
    else
    {
        if (localflag == 1)
        {
            localflag = 0;
            if (!spinflag)
                printtokenlist(tokenlistdefer, 4*indent, 1);
	    else
	    {
                ConvertToSpin(&tokenlistdefer, &tokenlistout, CLINE_STATEMENT);
	        tokenlistout = CheckTokenList(tokenlistout);
	        printtokenlist(tokenlistout, 2*indent, 1);
	        tokenlistout = ListDestroy(tokenlistout);
	        tokenlistdefer = ListDestroy(tokenlistdefer);
	    }
        }
        if (!spinflag)
            printtokenlist(tokenlist, 4*indent, 1);
        else
        {
            ConvertToSpin(&tokenlist, &tokenlistout, clinetype);
	    tokenlistout = CheckTokenList(tokenlistout);
	    printtokenlist(tokenlistout, 2*indent, newline);
	    tokenlistout = ListDestroy(tokenlistout);
        }
    }
    if (oneliner)
    {
        LinkT *endptr1 = GetEndPtr(1);
        if (endptr1)
        {
	    tokenlistout = ListDuplicate(endptr1);
	    printtokenlist(tokenlistout, 2*indent, newline);
	    tokenlistout = ListDestroy(tokenlistout);
	}
    }
    if (indent == 0 && clinetype == CLINE_BRACE_RIGHT)
    {
        if (debugflag) PrintVariableList(stdout);
        PurgeLocalParms();
        if (whitespace == 1)
            fprintf(outfile, "\n");
    }
}

void usage(void)
{
    printf("usage: cspin infile\n");
    exit(1);
}

void ParseCommandLine(int argc, char **argv)
{
    int i;

    if (argc < 2)
    {
	interact = 1;
	return;
    }

    inputfname[0][0] = 0;
    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            if (argv[i][1] == 'c')
                spinflag = 0;
            else if (argv[i][1] == 'd')
                debugflag = 1;
            else if (argv[i][1] == 'C')
                sourceflag = 1;
            else if (argv[i][1] == 'm')
                mainflag = 1;
            else if (argv[i][1] == 'i')
                interact = 1;
            else if (argv[i][1] == 'w')
                whitespace = argv[i][2] - '0';
            else
	        printf("Ignoring unknown option %s\n", argv[i]);
        }
        else
        {
            strncpy(inputfname[0], argv[i], 79);
            inputfname[0][79] = 0;
        }
    }
    if (inputfname[0][0])
        ConvertFileToSpin(inputfname[0]);
    else
        interact = 1;
}

void PrintHelp(char *str)
{
    if (*str) printf("%s\n", str);
    printf("\nCommand List\n");
    printf("help           - Print this information\n");
    printf("convert file   - Convert a C file to Spin\n");
    printf("dir            - List the files in the current working directory\n");
    printf("pwd            - Print the current working directory\n");
    printf("cd directory   - Change the working directory\n");
    printf("main on/off    - Enable/Disable the main mode (default off)\n");
    printf("spin on/off    - Enable/Disable generating spin code (default on)\n");
    printf("comment on/off - Include/Exclude C code as comments (default off)\n");
    printf("debug on/off   - Enable/Disable the debug output (default off)\n");
    printf("exit           - Exit cspin\n");
}

void CommandInterpreter(void)
{
    int fileopen = 0;
    char buffer[100];
    char *cmdstr;
    char *parmstr;
    char *endstr;
    DIR *pdir;
    struct dirent *pdirent;

    printf("\nCSPIN Version %s\n\n", versionstr);

    while (fileopen == 0)
    {
        printf("Enter command: ");
        gets(buffer);
	cmdstr = SkipChars(buffer, " \t");
	parmstr = FindChars(cmdstr, " \t");
	if (*parmstr) *parmstr++ = 0;
	parmstr = SkipChars(parmstr, " \t");
	endstr = FindChars(parmstr, " \t");
	*endstr = 0;
	if (strcmp(cmdstr, "help") == 0)
	    PrintHelp("");
	else if (strcmp(cmdstr, "exit") == 0)
	    exit(0);
	else if (strcmp(cmdstr, "convert") == 0)
	{
	    ConvertFileToSpin(parmstr);
	}
	else if (strcmp(cmdstr, "pwd") == 0)
	{
            getcwd(buffer, 100);
            printf("%s\n\n", buffer);
	}
	else if (strcmp(cmdstr, "dir") == 0)
	{
            getcwd(buffer, 100);
	    printf("\n%s\n", buffer);
	    pdir = opendir(".");
	    while ((pdirent = readdir(pdir))) printf("%s\n", pdirent->d_name);
	    printf("\n");
            closedir(pdir);
	}
	else if (strcmp(cmdstr, "cd") == 0)
	{
            chdir(parmstr);
            getcwd(buffer, 100);
            printf("%s\n\n", buffer);
	}
	else if (strcmp(cmdstr, "main") == 0)
	{
	    if (strcmp(parmstr, "on") == 0)
	        mainflag = 1;
	    else if (strcmp(parmstr, "off") == 0)
	        mainflag = 0;
	    else PrintHelp("Illegal Command");
	}
	else if (strcmp(cmdstr, "spin") == 0)
	{
	    if (strcmp(parmstr, "on") == 0)
	        spinflag = 1;
	    else if (strcmp(parmstr, "off") == 0)
	        spinflag = 0;
	    else PrintHelp("Illegal Command");
	}
	else if (strcmp(cmdstr, "comment") == 0)
	{
	    if (strcmp(parmstr, "on") == 0)
	        sourceflag = 1;
	    else if (strcmp(parmstr, "off") == 0)
	        sourceflag = 0;
	    else PrintHelp("Illegal Command");
	}
	else if (strcmp(cmdstr, "debug") == 0)
	{
	    if (strcmp(parmstr, "on") == 0)
	        debugflag = 1;
	    else if (strcmp(parmstr, "off") == 0)
	        debugflag = 0;
	    else PrintHelp("Illegal Command");
	}
        else if (*cmdstr) PrintHelp("Illegal Command");
    }
}

int main(int argc, char **argv)
{
    ParseCommandLine(argc, argv);

    if (interact)
        CommandInterpreter();

    return 0;
}


int ProcessIncludeDirective(LinkT *tokenlist)
{
    char *ptr;

    if (tokenlist == 0) return 0;
    if (strcmp(tokenlist->str, "#")) return 0;

    tokenlist = tokenlist->next;
    if (tokenlist == 0) return 0;
    if (strcmp(tokenlist->str, "include")) return 0;

    tokenlist = tokenlist->next;
    if (tokenlist == 0) return 0;
    if (tokenlist->str[0] != '"') return 0;

    if (inputfileindex > 9)
    {
	printf("Too many nested includes\n");
	return 0;
    }

    inputfileindex++;
    strncpy(inputfname[inputfileindex], tokenlist->str + 1, 79);
    inputfname[inputfileindex][79] = 0;
    ptr = FindChar(inputfname[inputfileindex], '"');
    *ptr = 0;
    inputline[inputfileindex] = 0;
    inputfile[inputfileindex] = fopen(inputfname[inputfileindex], "r");
    if (inputfile[inputfileindex] == 0)
    {
	inputfileindex--;
	printf("%s, %d - ", inputfname[inputfileindex], inputline[inputfileindex]);
	printf("Could not open include file %s\n", inputfname[inputfileindex+1]);
        printtokenlist(tokenlist, 0, 1);
    }

    return 1;
}

void InitializeTypesList(void)
{
    AddToTypesList("void", 1, 0);
    AddToTypesList("char", 1, charsign);
    AddToTypesList("short", 2, 1);
    AddToTypesList("int", 4, 1);
    AddToTypesList("long", 4, 1);
    AddToTypesList("float", 4, 1);
    AddToTypesList("double", 8, 1);
    AddToTypesList("FILE", 4, 1); /* Use this definition for now */
}

void AddIntrinsicFunctions(void)
{
    //AddIntrinsicFunction("int", "memset", "BYTEFILL");
    //AddIntrinsicFunction("int", "memcpy", "BYTEMOVE");
    //AddIntrinsicFunction("int", "strcmp", "STRCOMP");
    AddIntrinsicFunction("int", "strlen", "strsize");
}

void InitializeFuncProtoList(void)
{
    AddCFuncProtoToVarList("void", 1, "calloc", 2);
    AddCFuncProtoToVarList("int",  0, "chdir", 1);
    AddCFuncProtoToVarList("int",  0, "closedir", 1);
    AddCFuncProtoToVarList("int",  0, "create", 1);
    AddCFuncProtoToVarList("void", 0, "exit", 1);
    AddCFuncProtoToVarList("int",  0, "fclose", 1);
    AddCFuncProtoToVarList("int",  0, "fgetc", 1);
    AddCFuncProtoToVarList("char", 1, "fgets", 3);
    AddCFuncProtoToVarList("void", 1, "fopen", 2);
    AddCFuncProtoToVarList("int",  0, "format", 1);
    AddCFuncProtoToVarList("int",  0, "fputc", 2);
    AddCFuncProtoToVarList("int",  0, "fputs", 2);
    AddCFuncProtoToVarList("int",  0, "fread", 4);
    AddCFuncProtoToVarList("void", 0, "free", 1);
    AddCFuncProtoToVarList("int",  0, "fseek", 3);
    AddCFuncProtoToVarList("int",  0, "fwrite", 4);
    AddCFuncProtoToVarList("int",  0, "getch", 0);
    AddCFuncProtoToVarList("int",  0, "getchar", 0);
    AddCFuncProtoToVarList("char", 1, "getcwd", 0);
    AddCFuncProtoToVarList("char", 1, "gets", 1);
    AddCFuncProtoToVarList("int",  0, "isdigit", 1);
    AddCFuncProtoToVarList("int",  0, "tolower", 1);
    AddCFuncProtoToVarList("int",  0, "toupper", 1);
    AddCFuncProtoToVarList("int",  0, "itoa", 1);
    AddCFuncProtoToVarList("void", 1, "malloc", 1);
    AddCFuncProtoToVarList("void", 1, "memcpy", 3);
    AddCFuncProtoToVarList("void", 1, "memset", 3);
    AddCFuncProtoToVarList("int",  0, "mount", 1);
    AddCFuncProtoToVarList("void", 1, "opendir", 1);
    AddCFuncProtoToVarList("int",  0, "putchar", 1);
    AddCFuncProtoToVarList("int",  0, "puts", 1);
    AddCFuncProtoToVarList("void", 1, "readdir", 1);
    AddCFuncProtoToVarList("int",  0, "remove", 1);
    AddCFuncProtoToVarList("char", 1, "strcat", 2);
    AddCFuncProtoToVarList("int",  0, "strcmp", 2);
    AddCFuncProtoToVarList("char", 1, "strcpy", 2);
    AddCFuncProtoToVarList("int",  0, "strncmp", 3);
    AddCFuncProtoToVarList("char", 1, "strncpy", 3);
    AddCFuncProtoToVarList("int",  0, "vfprintf", 0);
    AddCFuncProtoToVarList("int",  0, "vprintf", 0);
    AddCFuncProtoToVarList("int",  0, "vscanf", 0);
    AddCFuncProtoToVarList("int",  0, "vsprintf", 0);
    AddCFuncProtoToVarList("int",  0, "vsscanf", 0);
    AddCFuncProtoToVarList("int",  0, "rand", 0);
    AddCFuncProtoToVarList("void", 0, "srand", 1);
    AddCFuncProtoToVarList("void", 0, "sleep", 1);
    AddCFuncProtoToVarList("void", 0, "usleep", 1);
}

void InitializeSymbolList(void)
{
    AddToSymbolList("and");
    AddToSymbolList("add");
    AddToSymbolList("byte");
    AddToSymbolList("call");
    AddToSymbolList("cmp");
    AddToSymbolList("cnt");
    AddToSymbolList("con");
    AddToSymbolList("enc");
    AddToSymbolList("fit");
    AddToSymbolList("not");
    AddToSymbolList("test");
    AddToSymbolList("or");
    AddToSymbolList("obj");
    AddToSymbolList("next");
    AddToSymbolList("sign");
    AddToSymbolList("pi");
    AddToSymbolList("c");
    AddToSymbolList("max");
    AddToSymbolList("min");
}

void PrintInitialInfo(void)
{
    fprintf(outfile, "\nCON\n");
    fprintf(outfile, "  _clkmode = xtal1 + pll16x\n");
    fprintf(outfile, "  _xinfreq = 5_000_000\n");
    fprintf(outfile, "\nOBJ\n");
    fprintf(outfile, "  c : \"clib\"\n");
    fprintf(outfile, "\nPUB start\n");
    fprintf(outfile, "  c.start\n");
    fprintf(outfile, "  cspin_initialize\n");
    fprintf(outfile, "  main\n");
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
