/*
############################################################################
# Copyright (c) 2010 - 2014 Dave Hein
# MIT Licensed
############################################################################
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "spinit.h"
#include "spinstruct.h"
#include "spinvars.h"
#include "cline.h"
#include "tokens.h"

extern VarInfoT *varlist;
extern VarInfoT *StructElem;

extern int modeflag;
extern TypeInfoT *typeslist;
extern TypeInfoT *structlink;
extern int nonewline;

void GenerateLocalSpinStructDeclaration(char *name, int size, LinkT **ptokenlistout)
{
    VarInfoT *varinfo = FindVariable(varlist, name);
    int i, numdims;
    int ptrlevel = 0;
    int arraysize = 0;

    if (varinfo)
    {
	name = varinfo->str1;
        numdims = varinfo->numdims;
	ptrlevel = varinfo->ptrlevel;
	if (numdims)
	{
            arraysize = 1;
            ptrlevel -= numdims;
            for (i = 0; i < numdims; i++)
                arraysize *= varinfo->dimsize[i];
	    if (ptrlevel > 0)
	        size = 4 * arraysize;
	    else
	        size *= arraysize;
	}
	else if (ptrlevel > 0)
	    size = 1;
    }

    AddToken(name, ptokenlistout);
    if (size != 1)
    {
        char *ptr = str_alloc(20);
        sprintf(ptr, "%d", size >> 2);
        AddToken("[", ptokenlistout);
        AddToken(ptr, ptokenlistout);
        AddToken("]", ptokenlistout);
    }
}

void GenerateGlobalSpinStructDeclaration(char *name, int size, LinkT **ptokenlistout)
{
    AddToken("  long", ptokenlistout);
    GenerateLocalSpinStructDeclaration(name, size, ptokenlistout);
}

void ConvertDeclStructToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    int size = 0;
    TypeInfoT *typeinfo;

    if (clinetype == CLINE_STRUCT_VAR)
    {
	i = 0;
        typeinfo = structlink;
    }
    else
    {
	i = 2;
        typeinfo= FindType(typeslist, tokens1[1]);
    }

    if (tokens1[i][0] == '*') i++;

    if (typeinfo) size = typeinfo->size;

    if (modeflag != MODEFLAG_VAR)
    {
        AddToken("\n", ptokenlistout);
        AddToken("VAR", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_VAR;
    }
    GenerateGlobalSpinStructDeclaration(tokens1[i], size, ptokenlistout);
    //if (!nonewline) AddToken("\n", ptokenlistout);
}

void ConvertStructToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    if (modeflag != MODEFLAG_CON)
    {
        AddToken("\n", ptokenlistout);
        AddToken("CON", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_CON;
    }
    AddToken("'", ptokenlistout);
    for (i = 0; i < num1; i++) AddToken(tokens1[i], ptokenlistout);
}

void ConvertStructElemToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    if (modeflag != MODEFLAG_CON)
    {
        AddToken("\n", ptokenlistout);
        AddToken("CON", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_CON;
    }
    for (i = 0; i < num1; i++) if (tokens1[i][0] == ';') break;
    if (i > 0 && i < num1)
    {
        char *ptr;
	if (StructElem) AddToken(StructElem->str1, ptokenlistout);
	AddToken("=", ptokenlistout);
	ptr = str_alloc(10);
	if (StructElem) sprintf(ptr, "%d", StructElem->value);
	else strcpy(ptr, "????");
	AddToken(ptr, ptokenlistout);
        //if (!nonewline) AddToken("\n", ptokenlistout);
    }
}

int LocateClosingBracket(int i, char **tokens1, int num1)
{
    int count;

    for (count = 1; i < num1; i++)
    {
        if (tokens1[i][0] == ']')
        {
            if (--count <= 0) break;
        }
        else if (tokens1[i][0] == '[')
        {
            count++;
        }
    }
    return i;
}

int ConvertArrayReferenceToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, VarInfoT *varinfo, int indirect, int dimsub)
{
    int j;
    int parenflag;
    int dimlevel = 0;
    TypeInfoT *typeinfo;
    int scale, numdims, ptrlevel;

    // Check the variable info pointer
    if (varinfo == 0) return i;

    // Get variable information
    numdims = varinfo->numdims;
    ptrlevel = varinfo->ptrlevel;
    typeinfo = varinfo->typeinfo;

    // Process array and struct references
    while (i < num1 - 2 && tokens1[i+1][0] == '[' && dimlevel < numdims - dimsub)
    {
        i += 2;
        dimlevel++;
        AddToken("+", ptokenlistout);

        if (ptrlevel > numdims)
            scale = 4;
        else
            scale = typeinfo->size;
        for (j = dimlevel; j < numdims; j++)
            scale *= varinfo->dimsize[j];
        if (scale != 1)
        {
            char tempstr[12];
            AddToken("(", ptokenlistout);
            sprintf(tempstr, "%d", scale);
            AddToken(tempstr, ptokenlistout);
            AddToken("*", ptokenlistout);
            parenflag = 1;
        }
        else
            parenflag = 0;

#if 0
        while (i < num1 && tokens1[i][0] != ']')
            AddToken(tokens1[i++], ptokenlistout);
#else
        j = LocateClosingBracket(i, tokens1, num1);
        if (j > i + 1)
            AddToken("(", ptokenlistout);
        ConvertStatementToSpin(i, tokens1, j, ptokenlistout, CLINE_STATEMENT);
        if (j > i + 1)
            AddToken(")", ptokenlistout);
        i = j;
#endif
        if (parenflag)
            AddToken(")", ptokenlistout);
    }
    return i;
}

int ConvertStructReferenceToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int indirect)
{
    //int j;
    //int num20 = num2;
    VarInfoT *varlink;
    TypeInfoT *typeinfo;
    VarInfoT *structvars;
    VarInfoT *varinfo;
    int ptrlevel;
    LinkT *token20;
    int refstructptr = 0;
    int noel = 1;

    varlink = FindVariable(varlist, tokens1[i]);
    if (varlink == 0) return i;
    varinfo = varlink;

    // Check if just struct name alone
    if (tokens1[i+1][0] != '[' &&
        tokens1[i+1][0] != '.' &&
	strcmp(tokens1[i+1], "->") != 0)
    {
        LinkT *link = ListTail(*ptokenlistout);
	AddToken(varinfo->str1, ptokenlistout);
        if (!varinfo->ptrlevel && (!link || strcmp(link->str, "@")))
	    AddToken("NOEL", ptokenlistout);
	return i;
    }

    //token20 = AddToken("????", ptokenlistout);
    token20 = AddToken("[", ptokenlistout);

    // Determine if pointer or not
#if 1
    ptrlevel = varlink->ptrlevel - varlink->numdims;
    if (ptrlevel < 1)
        AddToken("@", ptokenlistout);
#else
    j = i + 1;
    if (tokens1[j][0] == '[')
    {
	while (tokens1[j++][0] != ']')
	{
	    if (j >= num1) break;
	    j++;
	}
    }

    if (tokens1[j][0] == '.')
        AddToken("@", ptokenlistout);
#endif
    AddToken(varlink->str1, ptokenlistout);

    while (i < num1 - 2 && (tokens1[i+1][0] == '.' ||
        strcmp(tokens1[i+1], "->") == 0 || tokens1[i+1][0] == '['))
    {
	if (refstructptr)
	{
	    refstructptr = 0;
	    AddToken("]", ptokenlistout);
	    token20 = InsertToken("long", ptokenlistout, token20);
	    token20 = InsertToken("[", ptokenlistout, token20);
	}
        typeinfo = varlink->typeinfo;
        structvars = (VarInfoT *)typeinfo->structvars;
        if (structvars == 0)
        {
	    if (tokens1[i+1][0] != '[')
	    {
	        printf("**** NO STRUCT VARS ****\n");
	        AddToken("ERROR", ptokenlistout);
	    }
	    break;
        }
	if (tokens1[i+1][0] == '[')
	{
#if 0
	    char *str = str_alloc(15);
	    int parenflag = tokens1[i+3][0] != ']';
	    int size = typeinfo->size;
	    if (varlink->ptrlevel > 1) size = 4;
	    sprintf(str, "%d", size);
            AddToken("+", ptokenlistout);
            AddToken("(", ptokenlistout);
            AddToken(str, ptokenlistout);
            AddToken("*", ptokenlistout);
	    if (parenflag) AddToken("(", ptokenlistout);
	    i += 2;
	    while (i < num1 && tokens1[i][0] != ']')
                AddToken(tokens1[i++], ptokenlistout);
	    if (parenflag) AddToken(")", ptokenlistout);
            AddToken(")", ptokenlistout);
#else
            i = ConvertArrayReferenceToSpin(i, tokens1, num1, ptokenlistout, varlink, indirect, 0);
#endif
	    continue;
	}
        varinfo = FindVariable(structvars, tokens1[i+2]);
        if (varinfo == 0)
        {
	    PrintFileLine();
	    printf(" - **** %s NOT AN ELEMENT OF STRUCT %s ****\n",
	        tokens1[i+2], varlink->str);
	    CspinExit(1);
        }
        noel = 0;
        AddToken("+", ptokenlistout);
        AddToken(varinfo->str1, ptokenlistout);
	varlink = varinfo;
        i += 2;
	// Check if this is a struct pointer
	if (varinfo->ptrlevel < 1) continue;
	if (varinfo->ptrlevel < 2 && varinfo->numdims) continue;
	if (varinfo->typeinfo == 0) continue;
	if (varinfo->typeinfo->structvars == 0) continue;
	refstructptr = 1;
    }

    if (tokens1[i+1][0] == '[') indirect--;
    ptrlevel = varinfo->ptrlevel + indirect;
    if (ptrlevel > 0 && varinfo->numdims)
    {
        if (noel)
        {
            LinkT *link = ListFindPrev(*ptokenlistout, token20);
            if (link && !strcmp(link->str, "@"))
            {
                RemoveToken(ptokenlistout, link);
	        ReplaceToken("(", ptokenlistout, token20);
                AddToken(")", ptokenlistout);
            }
            else
            {
                AddToken("NOEL", ptokenlistout);
            }
        }
        else
        {
	    ReplaceToken("(", ptokenlistout, token20);
            AddToken(")", ptokenlistout);
        }
    }
    else
    {
	int leftval = IsLeftValue(i, tokens1, num1);
	char *ptr = GetSignExtensionString(varinfo, indirect, leftval);
        AddToken("]", ptokenlistout);
	token20 = InsertToken(GetSpinTypeString(varinfo, indirect), ptokenlistout, token20);
	if (ptr) token20 = InsertToken(ptr, ptokenlistout, token20);
    }
    return i;
}

int ConvertPointerReferenceToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int indirect)
{
    char *ptr;
    int j;
    int i0 = i;
    int leftval;
    int numdims = 0;
    int ptrlevel = 0;
    int arrayindex[15];
    int arrayindex1[15];
    int arraycount = 0;
    int bracketflag = 0;
    VarInfoT *varinfo = FindVariable(varlist, tokens1[i]);

    // Check the variable info pointer
    if (varinfo == 0)
    {
	AddToken(tokens1[i], ptokenlistout);
	return i;
    }

    // Locate and count the array indices
    for (i++; i < num1; i++)
    {
	if (bracketflag == 0)
	{
	    if (tokens1[i][0] != '[')
	    {
		i--;
		break;
	    }
	    bracketflag = 1;
	    arrayindex[arraycount] = i;
	}
	else if (tokens1[i][0] == '[') bracketflag++;
	else if (tokens1[i][0] == ']')
	{
	    if (--bracketflag > 0) continue;
	    arrayindex1[arraycount++] = i;
	}
    }

#if 0
    // Add an indirect index if -1
    if (indirect == -1 && arraycount < 2)
    {
	arrayindex[arraycount++] = -1;
    }
#else
    // Add dummy array entries for each level of indirection
    for (j = 0; j < -indirect && arraycount < 14; j++)
	arrayindex[arraycount++] = -1;
#endif

    // Determine the pointer level
    numdims = varinfo->numdims;
    ptrlevel = varinfo->ptrlevel;
    ptrlevel -= numdims;

    // Check if there are too many dereferences
    if (arraycount > numdims + ptrlevel)
    {
        PrintFileLine();
        printf(" - too many dereferences\n");
        printf("arraycount = %d, indirect = %d, numdims = %d, ptrlevel = %d\n",
            arraycount, indirect, numdims, ptrlevel);
        CspinExit(1);
    }

    // Handle using an array as a pointer by just referencing it by name 
    if (arraycount == 0)
    {
	if (varinfo->numdims) AddToken("@", ptokenlistout);
	AddToken(varinfo->str1, ptokenlistout);
	return i;
    }

    // Handle *ptr or ptr[] for a scalar pointer
    else if (ptrlevel && numdims == 0 && arraycount == 1)
    {
        leftval = IsLeftValue(i, tokens1, num1);
        ptr = GetSignExtensionString(varinfo, -arraycount, leftval);
        if (ptr) AddToken(ptr, ptokenlistout);
        AddToken(GetSpinTypeString(varinfo, -arraycount), ptokenlistout);
        AddToken("[", ptokenlistout);
        AddToken(varinfo->str1, ptokenlistout);
        i = i0;
        if (i < num1 - 1 &&
            (!strcmp(tokens1[i+1], "++") || !strcmp(tokens1[i+1], "--")))
            AddToken(tokens1[++i], ptokenlistout);
        AddToken("]", ptokenlistout);
    }

    // Handle a pointer array that is fully dereferenced
    else if (ptrlevel && (arraycount == numdims + ptrlevel))
    {
        int rawflag = (numdims == 1);

        leftval = IsLeftValue(i, tokens1, num1);
        ptr = GetSignExtensionString(varinfo, -arraycount, leftval);
        if (ptr) AddToken(ptr, ptokenlistout);
        AddToken(GetSpinTypeString(varinfo, -arraycount), ptokenlistout);
        AddToken("[", ptokenlistout);
        if (rawflag)
        {
            AddToken(varinfo->str1, ptokenlistout);
            if (arrayindex[0] > 0)
            {
                AddToken("[", ptokenlistout);
                ConvertStatementToSpin(arrayindex[0]+1, tokens1, arrayindex1[0], ptokenlistout, CLINE_STATEMENT);
	        i = arrayindex1[0];
                AddToken("]", ptokenlistout);
            }
        }
        else if (numdims)
        {
            AddToken("long", ptokenlistout);
            AddToken("[", ptokenlistout);
            AddToken("@", ptokenlistout);
            AddToken(varinfo->str1, ptokenlistout);
            i = ConvertArrayReferenceToSpin(i0, tokens1, num1, ptokenlistout, varinfo, indirect, 0);
            AddToken("]", ptokenlistout);
        }
        else
        {
            AddToken("long", ptokenlistout);
            AddToken("[", ptokenlistout);
            AddToken(varinfo->str1, ptokenlistout);
            AddToken("]", ptokenlistout);
            if (arrayindex[0] > 0)
            {
                AddToken("[", ptokenlistout);
                ConvertStatementToSpin(arrayindex[0]+1, tokens1, arrayindex1[0], ptokenlistout, CLINE_STATEMENT);
	        i0 = arrayindex1[0];
                AddToken("]", ptokenlistout);
            }
            i = i0;
        }
        AddToken("]", ptokenlistout);
    }

    // Handle a pointer array that is not fully dereferenced
    else if (ptrlevel)
    {
        int rawflag = (numdims <= 1);

        if (rawflag)
        {
            AddToken(varinfo->str1, ptokenlistout);
            if (arrayindex[0] > 0)
            {
                AddToken("[", ptokenlistout);
                ConvertStatementToSpin(arrayindex[0]+1, tokens1, arrayindex1[0], ptokenlistout, CLINE_STATEMENT);
	        i = arrayindex1[0];
                AddToken("]", ptokenlistout);
            }
        }
        else
        {
            AddToken("long", ptokenlistout);
            AddToken("[", ptokenlistout);
            AddToken("@", ptokenlistout);
            AddToken(varinfo->str1, ptokenlistout);
            i = ConvertArrayReferenceToSpin(i0, tokens1, num1, ptokenlistout, varinfo, indirect, 0);
            AddToken("]", ptokenlistout);
        }
    }

    // Handle a non-pointer array that is fully dereferenced
    else if (arraycount == numdims)
    {
	int rawflag = (numdims == 1) && (varinfo->use == CVARS_GLOBAL ||
            (varinfo->typeinfo && varinfo->typeinfo->size == 4));

        leftval = IsLeftValue(i, tokens1, num1);
        ptr = GetSignExtensionString(varinfo, -arraycount, leftval);
        if (ptr) AddToken(ptr, ptokenlistout);
        if (rawflag)
        {
            AddToken(varinfo->str1, ptokenlistout);
            if (arrayindex[0] > 0)
            {
                AddToken("[", ptokenlistout);
                ConvertStatementToSpin(arrayindex[0]+1, tokens1, arrayindex1[0], ptokenlistout, CLINE_STATEMENT);
	        i = arrayindex1[0];
                AddToken("]", ptokenlistout);
            }
        }
        else
        {
            AddToken(GetSpinTypeString(varinfo, -arraycount), ptokenlistout);
            AddToken("[", ptokenlistout);
            AddToken("@", ptokenlistout);
            AddToken(varinfo->str1, ptokenlistout);
            i = ConvertArrayReferenceToSpin(i0, tokens1, num1, ptokenlistout, varinfo, indirect, 1);
            AddToken("]", ptokenlistout);
        }
    }

    // Handle a non-pointer array that is not fully dereferenced
    else  // arraycount < numdims -- returns pointer
    {
        AddToken("(", ptokenlistout);
        AddToken("@", ptokenlistout);
        AddToken(varinfo->str1, ptokenlistout);
        i = ConvertArrayReferenceToSpin(i0, tokens1, num1, ptokenlistout, varinfo, indirect, 1);
        AddToken(")", ptokenlistout);
    }

    return i;
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
