/*
############################################################################
# Copyright (c) 2010 - 2014 Dave Hein
# MIT Licensed
############################################################################
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "spinit.h"
#include "spinstruct.h"
#include "spinvars.h"
#include "cline.h"
#include "tokens.h"
#include "cdirect.h"
#include "funcptr.h"

extern int nonewline;
static char *typelist[] = {"char", "short", "int", "long", "void", "FILE", "static", "extern", "unsigned", "signed", "float", "double", "volatile", 0};
extern int modeflag;
extern int prevtype;
extern TypeInfoT *typeslist;
extern VarInfoT *varlist;
extern ConInfoT *conlist;
extern int whitespace;
int decldatmode = 0;
static LinkT *datstrlist = 0;
static int datstroffset;
static char *datstrname;

char *GetSignExtensionString(VarInfoT *varlink, int indirect, int leftval)
{
    char *ptr = 0;
    int size = GetVarSizeLink(varlink, indirect);

    if (leftval == 0 && varlink->signflag == 1)
    {
        if (size == 1)
	    ptr = "~";
        else if (size == 2)
            ptr = "~~";
    }

    return ptr;
}

char *GetSpinTypeString(VarInfoT *varlink, int indirect)
{
    char *ptr;
    int size = GetVarSizeLink(varlink, indirect);

    if (size == 1)
        ptr = "byte";
    else if (size == 2)
        ptr = "word";
    else if (size == 4)
        ptr = "long";
    else if (size == 8)
        ptr = "long";
    else
        ptr = "long";

    return ptr;
}

void ConvertDirectiveToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    char *name;
    ConInfoT *coninfo;

    if (strcmp(tokens1[0], "#") == 0 && strcmp(tokens1[1], "macro") == 0) return;

#if 0
    if (prevtype == CLINE_FUNCTION || prevtype == CLINE_DECL_LOCAL)
        AddToken("\n", ptokenlistout);
#endif

    if (num1 > 0 && tokens1[num1-1][0] == '\n') num1--;

    if (num1 >= 3 && strcmp(tokens1[0], "#") == 0 && strcmp(tokens1[1], "define") == 0)
    {
	if (num1 < 4) return;

	name = tokens1[2];
        coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, name);

	if (coninfo == 0)
	{
	    PrintFileLine();
	    printf("Could not find %s in conlist\n", name);
	    return;
	}

        if (coninfo->type == CONTYPE_STRING) return;

        name = coninfo->str1;

        if (modeflag != MODEFLAG_CON)
        {
	    AddToken("\n", ptokenlistout);
	    AddToken("CON", ptokenlistout);
	    AddToken("\n", ptokenlistout);
	    modeflag = MODEFLAG_CON;
        }

        AddToken(" ", ptokenlistout);
        AddToken(name, ptokenlistout);
        AddToken("=", ptokenlistout);
	for (i = 3; i < num1; i++)
	{
	    if (tokens1[i][0] == '\n') break;
            AddToken(TranslateCStrToSpinStr(tokens1[i], 0), ptokenlistout);
	}
        //if (!nonewline) AddToken("\n", ptokenlistout);
    }
    else
    {
        AddToken("'CSPIN WARNING", ptokenlistout);
	for (i = 0; i < num1; i++)
            AddToken(tokens1[i], ptokenlistout);
#if 0
	if (tokens1[num1-1][0] != '\n')
            AddToken("\n", ptokenlistout);
#endif
    }
}

int SkipFuncPtrParms(int i, int num1, char **tokens1)
{
    int parencount;

    for (parencount = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '(')
            parencount++;
        else if (tokens1[i][0] == ')')
            parencount--;
        if (parencount <= 0) break;
    }
    return i;
}

int SkipBrackets(int i, int num1, char **tokens1)
{
    int bracketflag;

    for (bracketflag = 0; i < num1; i++)
    {
        if (bracketflag)
        {
            if (tokens1[i][0] == ']')
                bracketflag = 0;
        }
        else
        {
            if (tokens1[i][0] == '[')
                bracketflag = 1;
            else
                break;
        }
    }
    return i;
}

void AddArraySize(LinkT **ptokenlistout, VarInfoT *varinfo, int local)
{
    char tempstr[20];
    int size, elements, j;

    if (varinfo->ptrlevel == varinfo->numdims && varinfo->typeinfo)
        size = varinfo->typeinfo->size;
    else
        size = 4;

    elements = 1;
    for (j = 0; j < varinfo->numdims; j++)
        elements *= varinfo->dimsize[j];
    if (local)
    {
        if (size == 1) elements = (elements + 3) / 4;
        else if (size == 2) elements = (elements + 1) / 2;
    }
    sprintf(tempstr, "%d", elements);

    AddToken("[", ptokenlistout);
    AddToken(tempstr, ptokenlistout);
    AddToken("]", ptokenlistout);
}

int SkipBracketsAddArraySize(int i, int num1, char **tokens1, LinkT **ptokenlistout, VarInfoT *varinfo, int local)
{
    int size;
    int num2 = SkipBrackets(i, num1, tokens1);

    if (varinfo->ptrlevel == varinfo->numdims && varinfo->typeinfo)
        size = varinfo->typeinfo->size;
    else
        size = 4;

    if (varinfo->numdims == 1 && (!local || size == 4))
    {
        ConvertStatementToSpin(i, tokens1, num2, ptokenlistout, CLINE_STATEMENT);
    }
    else
    {
        AddArraySize(ptokenlistout, varinfo, local);
    }
    return num2;
}

void ConvertDeclLocalToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;
    int firstflag = 1;
    TypeInfoT *typeinfo;
    VarInfoT *varinfo;
    //int bracketflag;
    //int size, elements;
    //char tempstr[20];

    for (i = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '\n') break;
	if ((typeinfo = FindType(typeslist, tokens1[i])))
	{
	    if (typeinfo->structvars == 0 || tokens1[i+1][0] == '*') continue;
            if (firstflag)
            {
	        firstflag = 0;
	        if (prevtype == CLINE_FUNCTION)
	            AddToken("|", ptokenlistout);
	        else
	            AddToken(",", ptokenlistout);
            }
	    if (tokens1[i+1][0] == '*') i++;
            GenerateLocalSpinStructDeclaration(tokens1[i+1], typeinfo->size, ptokenlistout);
	    i++;
	    continue;
	}
        if (SearchList(tokens1[i], typelist, SEARCH_EXACT) >= 0) continue;
        if (tokens1[i][0] == '*') continue;
        if (tokens1[i][0] == ';') continue;
	if (tokens1[i][0] == ',')
	{
            AddToken(tokens1[i], ptokenlistout);
	    continue;
	}

        if (firstflag)
        {
	    firstflag = 0;
	    if (prevtype == CLINE_FUNCTION)
	        AddToken("|", ptokenlistout);
	    else
	        AddToken(",", ptokenlistout);
        }

	if ((varinfo = FindVariable(varlist, tokens1[i])) == 0) continue;

        AddToken(varinfo->str1, ptokenlistout);

        // Copy array info
        if (varinfo->numdims)
        {
            i = SkipBracketsAddArraySize(i+1, num1, tokens1, ptokenlistout, varinfo, 1) - 1;
        }

        if (varinfo->funcptr &&
            (!varinfo->typeinfo || !varinfo->typeinfo->funcptr))

            i = SkipFuncPtrParms(i+2, num1, tokens1) - 1;
    }
}

void ConvertDeclGlobalToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i;

    // Look for an equals sign
    for (i = 0; i < num1; i++)
    {
	if (strcmp(tokens1[i], "=") == 0) break;
    }

    if (i < num1 || decldatmode)
        ConvertDeclGlobalDatToSpin(tokens1, num1, ptokenlistout, clinetype);
    else
        ConvertDeclGlobalVarToSpin(tokens1, num1, ptokenlistout, clinetype);

}

void ConvertDeclGlobalVarToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i, k;
    int pointer = 0;
    VarInfoT *varinfo;
    TypeInfoT *typeinfo;
    TypeInfoT *typetemp;

    if (modeflag != MODEFLAG_VAR)
    {
        if (whitespace)
            AddToken("\n", ptokenlistout);
        AddToken("VAR", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_VAR;
    }

    typeinfo = FindType(typeslist, "int");
    k = 0;
    for (i = 0; i < num1; i++)
    {
        if (tokens1[i][0] == '*')
        {
	    pointer = 1;
	    continue;
	}
	else if ((typetemp = FindType(typeslist, tokens1[i])))
	{
	    typeinfo = typetemp;
	    if (typeinfo->structvars)
	    {
		if (tokens1[i+1][0] == '*') i++;
                GenerateGlobalSpinStructDeclaration(tokens1[i+1], typeinfo->size, ptokenlistout);
                //AddToken("\n", ptokenlistout);
		return;
	    }
	}
	else if (FindVariable(varlist, tokens1[i]))
	    break;
    }
    AddToken(" ", ptokenlistout);
    if (pointer || typeinfo->ptrlevel || typeinfo->funcptr)
        AddToken("long", ptokenlistout);
    else if (typeinfo->size == 1)
        AddToken("byte", ptokenlistout);
    else if (typeinfo->size == 2)
        AddToken("word", ptokenlistout);
    else if (typeinfo->size == 4)
        AddToken("long", ptokenlistout);
    else
        AddToken("????", ptokenlistout);

    for (; i < num1; i++)
    {
	if (tokens1[i][0] == ';') continue;
	if (tokens1[i][0] == ',')
	    AddToken(tokens1[i], ptokenlistout);
	else
        {
	    if ((varinfo = FindVariable(varlist, tokens1[i])))
            {
                AddToken(varinfo->str1, ptokenlistout);
                if (varinfo->numdims)
                {
                    i = SkipBracketsAddArraySize(i+1, num1, tokens1, ptokenlistout, varinfo, 0) - 1;
                }
                if (varinfo->funcptr &&
                    (!varinfo->typeinfo || !varinfo->typeinfo->funcptr))
                    i = SkipFuncPtrParms(i+2, num1, tokens1) - 1;
            }
            else
            {
	        AddToken("ERROR", ptokenlistout);
            }
	    //AddToken(GetSpinVariableName(tokens1[i]), ptokenlistout);
        }
    }

}


void ConvertDeclGlobalDatToSpin(char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int i = -1;
    int funcflag;
    VarInfoT *varinfo;
    int first = 1;

    // Set for DAT mode
    if (modeflag != MODEFLAG_DAT)
    {
        if (whitespace)
            AddToken("\n", ptokenlistout);
        AddToken("DAT", ptokenlistout);
        AddToken("\n", ptokenlistout);
        modeflag = MODEFLAG_DAT;
        decldatmode = 0;
    }

    if (decldatmode == 0)
    {
        // Locate the variable name
        for (i = 0; i < num1; i++)
        {
            if ((varinfo = FindVariable(varlist, tokens1[i]))) break;
        }

        // Check if variable found
        if (!varinfo)
        {
	    PrintFileLine();
            printf("Couldn't locate variable\n");
            return;
        }

        // Print Spin type and Spin variable name
        AddToken(" ", ptokenlistout);
        AddToken(varinfo->str1, ptokenlistout);
        AddToken(GetSpinTypeString(varinfo, -varinfo->numdims), ptokenlistout);
        decldatmode = GetVarSizeLink(varinfo, -varinfo->numdims);

        // Locate the equals sign
        for (i++; i < num1; i++)
        {
            if (tokens1[i][0] == '=') break;
        }
    }
    else
    {
        // Print Spin type and Spin variable name
        AddToken(" ", ptokenlistout);
        if (decldatmode == 1)
            AddToken("byte", ptokenlistout);
        else if (decldatmode == 2)
            AddToken("word", ptokenlistout);
        else
            AddToken("long", ptokenlistout);
    }

    // Copy data to the output
    for (i++; i < num1; i++)
    {
        if (tokens1[i][0] == ',')
        {
            if (i == num1 - 1) break;
            AddToken(",", ptokenlistout);
            continue;
        }
        if (tokens1[i][0] == '=') continue;
        if (tokens1[i][0] == '{') continue;
        if (tokens1[i][0] == '}') continue;
        if (tokens1[i][0] == ';')
        {
            decldatmode = 0;
            continue;
        }
        if (tokens1[i][0] == '\n') continue;
        if (tokens1[i][0] == '"')
        {
            //printf("Process %s\n", tokens1[i]);
            if (!datstrlist)
            {
                datstroffset = 16;
	        datstrname = AddToSymbolList("datstr");
                AddToken(" ", &datstrlist);
                AddToken(datstrname, &datstrlist);
                AddToken("byte", &datstrlist);
            }
            else if (first)
            {
                AddToken("\n", &datstrlist);
                AddToken(" ", &datstrlist);
                AddToken("byte", &datstrlist);
            }
            else
                AddToken(",", &datstrlist);
            first = 0;
            AddToken(tokens1[i], &datstrlist);
            AddToken(",", &datstrlist);
            AddToken("0", &datstrlist);
            AddToken("@", ptokenlistout);
            AddToken(datstrname, ptokenlistout);
            AddToken("+", ptokenlistout);
            AddTokenNum(datstroffset, ptokenlistout);
            datstroffset += strlen(tokens1[i]) - 1;
            continue;
        }
        if ((varinfo = FindVariable(varlist, tokens1[i])))
        {
            funcflag = (varinfo->use == CVARS_FUNCTION ||
                varinfo->use == CVARS_PROTOTYPE);
            if (varinfo->numdims || funcflag)
            {
                if (funcflag)
                {
                    AddToFuncPtrList(varinfo->str1, varinfo->numparms);
                    AddToken("@", ptokenlistout);
                    AddToken(GetFuncPtrName(varinfo->str1), ptokenlistout);
                }
                else
                {
                    AddToken("@", ptokenlistout);
                    AddToken(varinfo->str1, ptokenlistout);
                }
                AddToken("+", ptokenlistout);
                AddToken("16", ptokenlistout);
                continue;
            }
        }
        AddToken(TranslateCStrToSpinStr(tokens1[i], 0), ptokenlistout);
    }
    if (!decldatmode && datstrlist)
    {
        LinkT *tail = AddToken("\n", ptokenlistout);
        tail->next = datstrlist;
        datstrlist = 0;
    }
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
